#!/usr/bin/env bash

# Import functions and environment variables
LIMINE_FUNCTIONS_PATH=/usr/lib/limine/limine-common-functions
# shellcheck disable=SC1090
source "${LIMINE_FUNCTIONS_PATH}" || {
    echo -e "\033[1;31m Error: Failed to source '${LIMINE_FUNCTIONS_PATH}'. Ensure the file exists.\033[0m"
    exit 1
}

initialize_header || exit 1

# Function to add Limine to the EFI boot entries
add_efi_entry() {
    local esp_path="$1"
    local boot_label="$2"
    local efi_file_path="$3"

    # Get the disk and partition information
    local source
    source=$(findmnt -n -o SOURCE "$esp_path") || {
        echo -e "\033[1;31m Error: Unable to determine the source device for '$esp_path'.\033[0m"
        return 1
    }

    # Check if the boot entry already exists
    if efibootmgr | grep -Fqi "/${efi_file_path}"; then
        #echo "Info: EFI boot entry '$boot_label' already exists. No action needed."
        return 0
    fi

    # Extract disk and partition
    local disk part
    if [[ "$source" =~ ^(/dev/nvme[0-9]+n[0-9]+)(p[0-9]+)$ ]]; then
        # NVMe device: /dev/nvmeXnYpZ -> disk=/dev/nvmeXnY, part=Z
        disk="${BASH_REMATCH[1]}"
        part="${BASH_REMATCH[2]#p}" # Remove leading 'p'
    elif [[ "$source" =~ ^(/dev/[a-z]+)([0-9]+)$ ]]; then
        # Other devices: /dev/sda1 -> disk=/dev/sda, part=1
        disk="${BASH_REMATCH[1]}"
        part="${BASH_REMATCH[2]}"
    else
        echo -e "\033[1;31m Error: Failed to parse disk and partition from source '$source'.\033[0m"
        return 1
    fi

    # Add the EFI entry using efibootmgr
    if efibootmgr --create \
        --disk "${disk}" \
        --part "${part}" \
        --label "${boot_label}" \
        --loader "${efi_file_path}" \
        --unicode; then
        echo "Success: EFI boot entry '${boot_label}' for '${efi_file_path}' added successfully."
    else
        echo -e "\033[1;31m Error: Failed to add EFI boot entry: disk=${disk}, part=${part}, label=${boot_label}, loader=${efi_file_path}\033[0m"
        return 1
    fi
}

# Function to update a specific EFI file if needed
update_efi_file() {
    local source_file="$1"
    local backup_file="$2"
    local target_file="$3"

    if [[ -f "$backup_file" ]]; then
        local source_hash
        local backup_hash
        # Calculate hashes
        source_hash=$(b2sum "$source_file" | awk '{print $1}')
        backup_hash=$(b2sum "$backup_file" | awk '{print $1}')
        if [[ "$source_hash" == "$backup_hash" ]]; then
            #echo "No update needed for $target_file (identical hash)."
            return 0
        fi
    fi

    echo "Updating $target_file..."
    if ! cp "$source_file" "$backup_file"; then
        echo -e "\033[1;31mError: Failed to copy $source_file to $backup_file.\033[0m"
        return 1
    fi
    if ! cp "$backup_file" "$target_file"; then
        echo -e "\033[1;31mError: Failed to copy $backup_file to $target_file.\033[0m"
        return 1
    fi
}

# Validate ESP_PATH
# shellcheck disable=SC2153
if [[ ! -d "${ESP_PATH}" ]]; then
    echo -e "\033[1;31mError: '${ESP_PATH}' is invalid. Please set ESP_PATH correctly.\033[0m"
    exit 1
fi

# Verify the original Limine binary file exists
if [[ ! -f "${BINARY_SOURCE_PATH}" ]]; then
    echo -e "\033[1;31mError: Limine EFI file '${BINARY_SOURCE_PATH}' is not available!\033[0m"
    exit 1
fi

# Create necessary directories
if ! install -dm 700 "${ESP_PATH}/EFI/BOOT/" "${ESP_PATH}/EFI/Limine/"; then
    echo -e "\033[1;31mError: Failed to create directories in '${ESP_PATH}'.\033[0m"
    exit 1
fi

# Optionally add some bootloaders if they are present
if [[ "${FIND_BOOTLOADERS}" == "yes" ]]; then
    if [[ -f "${ESP_PATH}/EFI/Microsoft/Boot/bootmgfw.efi" ]]; then
        limine-entry-tool --add-efi "Windows" "${ESP_PATH}/EFI/Microsoft/Boot/bootmgfw.efi" --comment "Microsoft bootloader" --priority 20 --quiet
    fi
    if [[ -f "${ESP_PATH}/EFI/systemd/systemd-bootx64.efi" ]]; then
        limine-entry-tool --add-efi "Systemd-boot" "${ESP_PATH}/EFI/systemd/systemd-bootx64.efi" --comment "Systemd bootloader" --priority 20 --quiet
    fi
    if [[ -f "${ESP_PATH}/EFI/refind/refind_x64.efi" ]]; then
        limine-entry-tool --add-efi "rEFInd" "${ESP_PATH}/EFI/refind/refind_x64.efi" --comment "rEFInd bootloader" --priority 20 --quiet
    fi
    if [[ -f "${ESP_PATH}/EFI/BOOT/BOOTX64.EFI" ]]; then
        limine-entry-tool --add-efi "EFI fallback" "${ESP_PATH}/EFI/BOOT/BOOTX64.EFI" --comment "Default EFI loader" --priority 10 --quiet
    fi
fi

mutex_lock "limine-efi-deploy"
# Update Limine EFI file
if ! update_efi_file "${BINARY_SOURCE_PATH}" "${BINARY_BACKUP_PATH}" "${BINARY_TARGET_PATH}"; then
    mutex_unlock
    exit 1
fi
enroll_config

# Optionally set Limine as fallback bootloader
if [[ "${SET_LIMINE_AS_FALLBACK}" == "yes" ]]; then
    cp "${BINARY_BACKUP_PATH}" "${BINARY_FALLBACK_PATH}"
fi

# Optionally register Limine in EFI boot entries
if [[ "${REGISTER_LIMINE_EFI}" == "yes" ]]; then
    if ! add_efi_entry "${ESP_PATH}" "Limine" "${LIMINE_EFI_PATH}"; then
        mutex_unlock
        exit 1
    fi
fi

mutex_unlock
echo "Limine EFI update completed successfully."
